"""Airbrake-Python package attributes and metadata."""

__all__ = (
    '__title__',
    '__summary__',
    '__author__',
    '__email__',
    '__version__',
    '__keywords__',
    '__license__',
    '__url__',
    '__notifier__',
)

__title__ = 'airbrake'
__summary__ = 'Python SDK for airbrake.io'
__author__ = 'BK Box, Sam Stavinoha'
__email__ = 'samuel.stavinoha@rackspace.com'
__version__ = '1.3.3'
__keywords__ = ['airbrake', 'exceptions', 'airbrake.io']
__license__ = 'Apache License, Version 2.0'
__url__ = 'https://github.com/airbrake/airbrake-python'
__notifier__ = {
    'name': 'airbrake-python',
    'version': __version__,
    'url': __url__,
}



"""Airbrake notifier client module.

Initialize this class to ship errors to airbrake.io
using the log() method.
"""

import json
import os
import platform
import socket
import sys
import traceback

import requests

from airbrake.__about__ import __notifier__
from airbrake import exc as ab_exc
from airbrake import utils


class Airbrake(object):
    """
    Class to handle exceptions, format & ship to airbrake.io.

    The payload (dict) to be POSTed as json is defined by an
    instance's `payload` attribute, which can be modified after
    Airbrake instantiation.

    Errors logged ( using Airbrake.log() ) will automatically be shipped.

    Usage *without* AirbrakeHandler:

        import airbrake
        ab = Airbrake(project_id="1234", api_key="1234")

        try:
            1/0
        except Exception:
            ab.log()

    The airbrake.io docs used to implement this class are here:
        http://help.airbrake.io/kb/api-2/notifier-api-v3
    """

    def __init__(self, project_id=None, api_key=None, environment=None,
                 base_url=None):
        """Client constructor."""
        # properties
        self._api_url = None
        self._context = None
        self.deploy_url = "http://api.airbrake.io/deploys.txt"
        self.notifier = __notifier__

        if not environment:
            environment = (os.getenv('AIRBRAKE_ENVIRONMENT') or
                           socket.gethostname())
        if not project_id:
            project_id = os.getenv('AIRBRAKE_PROJECT_ID', '')
        if not api_key:
            api_key = os.getenv('AIRBRAKE_API_KEY', '')
        if not base_url:
            base_url = os.getenv('AIRBRAKE_BASE_URL',
                                 'https://airbrake.io').strip('/')

        self.environment = str(environment)
        self.project_id = str(project_id)
        self.api_key = str(api_key)
        self.base_url = str(base_url)

        if not all((self.project_id, self.api_key)):
            raise ab_exc.AirbrakeNotConfigured(
                "Airbrake API Key (api_key) and Project ID "
                "(project_id) must be set. These values "
                "may be set using the environment variables "
                "AIRBRAKE_API_KEY and AIRBRAKE_PROJECT_ID or "
                "by passing in the arguments explicitly."
            )

        self._exc_queue = utils.CheckableQueue()

    def __repr__(self):
        """Return value for the repr function."""
        return ("Airbrake(project_id=%s, api_key=*****, environment=%s)"
                % (self.project_id, self.environment))

    @property
    def context(self):
        """The python, os, and app environment context."""
        if not self._context:
            self._context = {}
            # python
            version = platform.python_version()
            self._context.update({'language': 'Python %s' % version})
            # os
            plat = platform.platform()
            self._context.update({'os': plat})
            # env name
            self._context.update({'environment': self.environment})
            # TODO(samstav)
            #   add user info:
            #       userID, userName, userEmail
            #   add application info:
            #       version, url, rootDirectory
        return self._context

    @property
    def api_url(self):
        """Create the airbrake api endpoint and return a string."""
        if not self._api_url:
            self._api_url = "%s/api/v3/projects/%s/notices" % (
                self.base_url, self.project_id)
        return self._api_url

    def log(self, exc_info=None, message=None, filename=None,
            line=None, function=None, errtype=None, **params):
        """Acknowledge an error and post it to airbrake.io.

        :param exc_info:    Exception tuple to use for formatting request.
                            If None, sys.exc_info() will be read to get
                            exception info. To prevent the reading of
                            sys.exc_info(), set exc_info to False.
        :param message:     Message accompanying error.
        :param filename:    Name of file where error occurred.
        :param line:        Line number in file where error occurred.
        :param function:    Function name where error occurred.
        :param errtype:     Type of error which occurred.
        :param params:      Payload field "params" which may contain any other
                            context related to the exception(s).

        Exception info willl be read from sys.exc_info() if it is not
        supplied. To prevent this behavior, pass exc_info=False.
        """
        if not utils.is_exc_info_tuple(exc_info):
            # compatibility, allows exc_info not to be a exc tuple
            errmessage = None
            if isinstance(exc_info, Exception):
                errmessage = utils.pytb_lastline(exc_info)
            elif exc_info:
                try:
                    errmessage = json.dumps(exc_info)
                except (ValueError, TypeError):
                    errmessage = str(exc_info)
            if errmessage:
                # this way, if exc_info kwarg is passed to logger method,
                # its value will be available in params
                params['exc_info'] = errmessage

            if message and errmessage and message != errmessage:
                message = "%s | %s" % (message, errmessage)
            elif errmessage:
                message = errmessage

            # read the global exception stack, but
            # be sure not to read the same exception twice.
            # This can happen for a number of reasons:
            #   - the exception context has not been
            #     cleared or garbage collected
            #   - the airbrake logging handler is being used
            #     to log exceptions and error messages in the same
            #     thread, but that thread is not actually logging
            #     some exceptions, or is not clearing the exception
            #     context via sys.exc_clear()
            # the alternative is to clear the exception stack every time,
            # which this module *should not* take the liberty to do
            #
            # to prevent this function from reading the global
            # exception context *at all*, pass exc_info=False
            if exc_info is False:
                exc_info = None, None, None
            else:
                exc_info = sys.exc_info()

        # dont ship the same exception instance twice
        if exc_info[1] in self._exc_queue:
            exc_info = None, None, None
        self._exc_queue.put(exc_info[1])

        error = Error(
            exc_info=exc_info, message=message, filename=filename,
            line=line, function=function, errtype=errtype)
        environment = params.pop('environment', {})
        session = params.pop('session', {})
        payload = {'context': self.context,
                   'params': params,
                   'errors': [error.data],
                   'notifier': self.notifier,
                   'environment': environment,
                   'session': session}
        return self.notify(payload)

    def notify(self, payload):
        """Post the current errors payload body to airbrake.io.

        :param dict payload:
            Notification payload, in a dict/object form. The notification
            payload will ultimately be sent as a JSON-encoded string, but here
            it still needs to be in object form.
        """
        headers = {'Content-Type': 'application/json'}
        api_key = {'key': self.api_key}
        response = requests.post(
            self.api_url,
            data=json.dumps(payload, cls=utils.FailProofJSONEncoder,
                            sort_keys=True),
            headers=headers,
            params=api_key)
        response.raise_for_status()
        return response

    def deploy(self, env=None):
        """Reset counted errors for the airbrake project/environment."""
        environment = env or self.environment

        params = {'api_key': self.api_key,
                  'deploy[rails_env]': str(environment)}

        response = requests.post(self.deploy_url, params=params)
        response.raise_for_status()
        return response


class Error(object):  # pylint: disable=too-few-public-methods
    """Format the exception according to airbrake.io v3 API.

    The airbrake.io docs used to implements this class are here:
        http://help.airbrake.io/kb/api-2/notifier-api-v3
    """

    def __init__(self, exc_info=None, message=None, filename=None,
                 line=None, function=None, errtype=None):
        """Error object constructor."""
        self.exc_info = exc_info

        # default datastructure
        self.data = {
            'type': errtype or "Record",
            'backtrace': [{'file': filename or "N/A",
                           'line': line or 1,
                           'function': function or "N/A"}],
            'message': message or "N/A"}

        if utils.is_exc_info_tuple(self.exc_info):
            if not all(self.exc_info):
                return
            tbmessage = utils.pytb_lastline(self.exc_info)
            self.data.update(
                {'type': self.exc_info[1].__class__.__name__,
                 'backtrace': format_backtrace(self.exc_info[2]),
                 'message': message or tbmessage})
        else:
            raise TypeError(
                "Airbrake module (notifier.Error) received "
                "unsupported 'exc_info' type. Should be a "
                "3-piece tuple as returned by sys.exc_info(). "
                "Invalid argument was %s"
                % self.exc_info)


def format_backtrace(trace):
    """Create a formatted dictionary from a traceback object."""
    backtrace = []
    for filename, line, func, _ in traceback.extract_tb(trace):
        desc = {'file': filename,
                'line': line,
                'function': func}
        backtrace.append(desc)
    return backtrace



"""airbrake-python.

Python SDK for airbrake.io
"""

# pylint: disable=wildcard-import

import inspect
import logging
import os

from airbrake.__about__ import *  # noqa
from airbrake import exc  # noqa
from airbrake.notifier import Airbrake  # noqa
from airbrake.handler import AirbrakeHandler  # noqa


def getLogger(name=None, **kwargs):  # pylint: disable=invalid-name
    """Return a Logger with an AirbrakeHandler."""
    if not name:
        curframe = inspect.currentframe()
        callingpath = inspect.getouterframes(curframe, 2)[1][1]
        name = os.path.split(
            callingpath.rpartition('.')[0] or callingpath)[-1]
        name = "%s%s" % ('airbrake-python-', name)
    logger = logging.getLogger(name)

    if not has_airbrake_handler(logger):
        abh = AirbrakeHandler(**kwargs)
        logger.addHandler(abh)
        if logger.getEffectiveLevel() == logging.NOTSET:
            logger.setLevel(abh.level)
        elif not logger.isEnabledFor(abh.level):
            logger.setLevel(abh.level)

    return logger


def has_airbrake_handler(logger):
    """Check a logger for an AirbrakeHandler."""
    return any([isinstance(handler, AirbrakeHandler)
                for handler in logger.handlers])



"""Airbrake-Python custom exceptions."""


class AirbrakeException(Exception):

    """Base class for errors thrown by airbrake-python."""


class AirbrakeNotConfigured(AirbrakeException):

    """The client was not given id and key nor found in env."""



"""Airbrake-Python packaging and installation."""

import os
import subprocess
import sys

from setuptools import find_packages
from setuptools import setup


src_dir = os.path.dirname(os.path.realpath(__file__))
about = {}
with open(os.path.join(src_dir, 'airbrake', '__about__.py')) as abt:
    exec(abt.read(), about)  # pylint: disable=exec-used


# README.rst is for PyPI page
# pandoc --from=markdown_github --to=rst README.md --output=README.rst
LONG_DESCRIPTION = ''
readme_rst = os.path.join(src_dir, 'README.rst')
if os.path.isfile(readme_rst):
    with open(readme_rst) as rdme:
        LONG_DESCRIPTION = rdme.read()


# Add the commit hash to the keywords for sanity.
if any(k in ' '.join(sys.argv).lower() for k in ['upload', 'dist']):
    try:
        current_commit = subprocess.check_output(
            ['git', 'rev-parse', 'HEAD']).strip()
    except (OSError, subprocess.CalledProcessError):
        pass
    else:
        if current_commit and len(current_commit) == 40:
            about['__keywords__'].append(current_commit[:8])


CLASSIFIERS = [
    'Intended Audience :: Developers',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Topic :: Software Development',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: Implementation :: CPython',
]


INSTALL_REQUIRES = [
    'requests>=2.8.1'
]


TESTS_REQUIRE = [
    'coverage>=4.0',
    'flake8>=2.4.1',
    'flake8-docstrings>=0.2.1.post1',
    'mock>=1.3.0',
    'nose>=1.3.7',
    'pylint>=1.4.4',
    'testfixtures>=4.3.3',
]


package_attributes = {
    'author': about['__author__'],
    'author_email': about['__email__'],
    'classifiers': CLASSIFIERS,
    'description': about['__summary__'],
    'install_requires': INSTALL_REQUIRES,
    'keywords': ' '.join(about['__keywords__']),
    'license': about['__license__'],
    'long_description': LONG_DESCRIPTION,
    'name': about['__title__'],
    'packages': find_packages(exclude=['tests']),
    'test_suite': 'tests',
    'url': about['__url__'],
    'version': about['__version__'],
}

setup(**package_attributes)



"""Util functions/classes for airbrake-python."""

try:
    from queue import Queue, Full, Empty
except ImportError:
    # Py2 legacy fix
    from Queue import Queue, Full, Empty

import traceback
import types
import json

try:
    TypeType = types.TypeType
except AttributeError:
    # For >= Python 3
    TypeType = type


class FailProofJSONEncoder(json.JSONEncoder):
    """Uses object's representation for unsupported types."""

    def default(self, o):  # pylint: disable=E0202
        # E0202 ignored in favor of compliance with documentation:
        # https://docs.python.org/2/library/json.html#json.JSONEncoder.default
        """Return object's repr when not JSON serializable."""
        try:
            return repr(o)
        except Exception:  # pylint: disable=W0703
            return super(FailProofJSONEncoder, self).default(o)


class CheckableQueue(Queue):

    """Checkable FIFO Queue which makes room for new items."""

    def __init__(self, maxsize=1000):
        """Queue constructor."""
        Queue.__init__(self, maxsize=maxsize)

    def __contains__(self, item):
        """Check the Queue for the existence of an item."""
        try:
            with self.mutex:
                return item in self.queue
        except AttributeError:
            return item in self.queue

    def put(self, item, block=False, timeout=1):
        """Add an item to the Queue."""
        try:
            Queue.put(self, item, block=block, timeout=timeout)
        except Full:
            try:
                self.get_nowait()
            except Empty:
                pass
            self.put(item)


def is_exc_info_tuple(exc_info):
    """Determine whether 'exc_info' is an exc_info tuple.

    Note: exc_info tuple means a tuple of exception related values
    as returned by sys.exc_info().
    """
    try:
        errtype, value, tback = exc_info
        if all([x is None for x in exc_info]):
            return True
        elif all((isinstance(errtype, TypeType),
                  isinstance(value, Exception),
                  hasattr(tback, 'tb_frame'),
                  hasattr(tback, 'tb_lineno'),
                  hasattr(tback, 'tb_next'))):
            return True
    except (TypeError, ValueError):
        pass
    return False


def pytb_lastline(excinfo=None):
    """Return the actual last line of the (current) traceback.

    To provide exc_info, rather than allowing this function
    to read the stack frame automatically, this function
    may be called like so:

        ll = pytb_lastline(sys.exc_info())

    OR
        try:
            1/0
        except Exception as err:
            ll = pytb_lastline(err)
    """
    lines = None
    if excinfo:
        if isinstance(excinfo, Exception):
            kls = getattr(excinfo, '__class__', '')
            if kls:
                kls = str(getattr(kls, '__name__', ''))
                kls = ("%s: " % kls) if kls else ''
            lines = [kls + str(excinfo)]
        else:
            try:
                lines = traceback.format_exception(*excinfo)
                lines = "\n".join(lines).split('\n')
            except (TypeError, AttributeError) as err:
                err.message = (
                    "Incorrect argument(s) [%s] passed to pytb_lastline(). "
                    "Should be sys.exc_info() or equivalent. | %s"
                    % (excinfo, err.message))
                raise err.__class__(err.message)
    if not lines:
        # uses sys.exc_info()
        lines = traceback.format_exc().split('\n')
    # strip whitespace, Falsy values,
    # and the string 'None', sometimes returned by the traceback module
    lines = [line.strip() for line in lines if line]
    lines = [line for line in lines if str(line).lower() != 'none']
    if lines:
        return lines[-1]



"""AirbrakeHandler module.

All functions and types related to python logging
should be defined in this module.

A function for mapping a LogRecord object
    https://docs.python.org/2/library/logging.html#logrecord-objects
to an Airbrake error should be included here.
"""
import logging

from airbrake.notifier import Airbrake

_FAKE_LOGRECORD = logging.LogRecord('', '', '', '', '', '', '', '')


class AirbrakeHandler(logging.Handler):
    """A handler class which ships logs to airbrake.io.

    Requires one:
        * `project_id` AND `api_key`
        * an instance of airbrake.Airbrake
    """

    def __init__(self, airbrake=None, level=logging.ERROR, project_id=None,
                 api_key=None, environment=None, base_url=None):
        """Initialize the Airbrake handler with a default logging level.

        Default level of logs handled by this class are >= ERROR,
        which includes ERROR and CRITICAL. To change this behavior
        supply a different argument for 'level'.
        """

        logging.Handler.__init__(self, level=level)

        if isinstance(airbrake, Airbrake):
            self.airbrake = airbrake
        else:
            self.airbrake = Airbrake(project_id, api_key, environment,
                                     base_url)

    def emit(self, record):
        """Log the record airbrake.io style.

        To prevent method calls which invoke this handler from using the
        global exception context in sys.exc_info(), exc_info must be passed
        as False.

        E.g. To prevent AirbrakeHandler from reading the global exception
        context, (which it may do to find a traceback and error type), make
        logger method calls like this:

            LOG.error("Bad math.", exc_info=False)

        Otherwise, provide exception context directly, though the following
        contrived example would be a strange way to use the handler:

            exc_info = sys.exc_info()
            ...
            LOG.error("Bad math.", exc_info=exc_info)
        """
        try:
            airbrakeerror = airbrake_error_from_logrecord(record)
            self.airbrake.log(**airbrakeerror)
        except (KeyboardInterrupt, SystemExit):
            raise
        except:  # pylint: disable=bare-except
            self.handleError(record)


def airbrake_error_from_logrecord(record):
    """Create an airbrake error dictionary from a python LogRecord object.

    For more info on the logging.LogRecord class:
        https://docs.python.org/2/library/logging.html#logrecord-objects
    """
    airbrakeerror = {}
    params = {
        'logrecord_filename': record.filename,
        'levelname': record.levelname,
        'created': record.created,  # TODO(samstav): make this human readable
        'process_id': record.process,
        'process_name': record.processName,
        'thread_name': record.threadName,
        'lineno': record.lineno,
        'pathname': record.pathname,
        'funcName': record.funcName,
        'msg': record.getMessage()}

    # find params from kwarg 'extra'
    # See "The second keyword argument is extra"
    #   - https://docs.python.org/2/library/logging.html#logging.Logger.debug
    for key, val in list(vars(record).items()):
        if not hasattr(_FAKE_LOGRECORD, key):
            # handle attribute/field name collisions:
            # logrecod attrs should not limit or take
            # precedence over values specified in 'extra'
            if key in params:
                # if key "already" in params -> is logrecord attr
                params["logrecord_" + key] = params.pop(key)
            # let 'extra' (explicit) take precedence
            params[key] = val

    airbrakeerror.update(params)
    # errtype may be overridden by the notifier if an applicable
    # exception context is available and exc_info is not False
    airbrakeerror['errtype'] = "%s:%s" % (record.levelname, record.filename)
    airbrakeerror['exc_info'] = record.exc_info
    airbrakeerror['message'] = record.getMessage()
    airbrakeerror['filename'] = record.pathname
    airbrakeerror['line'] = record.lineno
    airbrakeerror['function'] = record.funcName
    return airbrakeerror
